from configparser import ConfigParser
import os
from .services import LighthouseService, GreenWebService
from .statistics import PageStatistics, StatisticsBuilder
from .exceptions import CarbonCalculatorException
import validators
from datetime import date, datetime, timezone
import json


class CarbonCalculator(object):
    """Carbon Calculator - The main module:"""

    def __init__(
        self, greenweb: GreenWebService, lighthouse: LighthouseService
    ) -> None:
        self._builder = StatisticsBuilder(greenweb, lighthouse)
        self._statistics = None

    def footprint(self, url: str) -> dict:
        if not validators.url(url):
            raise CarbonCalculatorException("The URL is not valid")

        try:
            self._statistics = None
            self._statistics = self._builder.build(url)
            self._co2 = self._statistics
        except Exception as e:
            raise CarbonCalculatorException(e)

    @property
    def date(self):
        return self._statistics.created_at

    @property
    def url(self) -> str:
        return self._statistics.url

    @property
    def hosting_green(self) -> bool:
        return self._statistics.hosting_green

    @property
    def co2_grams(self) -> float:
        return self._statistics.co2_grams

    @property
    def energy_kWh(self) -> float:
        return self._statistics.energy_kWh

    @property
    def water_litres(self) -> float:
        return self._statistics.water_litres

    @property
    def transfer_size_bytes(self) -> float:
        return self._statistics.transfer_size_bytes

    @property
    def resources_size_bytes(self) -> float:
        return self._statistics.resources_size_bytes

    @property
    def resources(self) -> dict:
        return self._statistics.resources

    def to_dict(self) -> dict:
        output = {}
        output["date"] = self._statistics.created_at
        output["url"] = self._statistics.url
        output["hosting_green"] = self._statistics.hosting_green
        output["co2_grams"] = self._statistics.co2_grams
        output["energy_kWh"] = self._statistics.energy_kWh
        output["water_litres"] = self._statistics.water_litres
        output["resources"] = self._statistics.resources

        return output

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), indent=4, default=json_serial)

    @classmethod
    def from_ini_file(cls, config_file: str) -> None:
        config = ConfigParser()
        config.read(config_file)

        lighthouse_filename = ""
        if config["LIGHTHOUSE"]["LIGHTHOUSE_PATH"] != "":
            lighthouse_filename = os.path.join(
                os.path.dirname(__file__),
                "../../",
                config["LIGHTHOUSE"]["LIGHTHOUSE_PATH"],
            )

        greenweb_filename = os.path.join(
            os.path.dirname(__file__), "../../", config["GREENWEB"]["GREENWEB_PATH"]
        )

        lighthouse = LighthouseService(lighthouse_filename)
        greenweb = GreenWebService(greenweb_filename)

        return cls(lighthouse=lighthouse, greenweb=greenweb)


def json_serial(obj):
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))



import argparse
from carbon.calculator import CarbonCalculator
from carbon.services import LighthouseService, GreenWebService


def main():
    """ """
    parser = argparse.ArgumentParser(
        description="Carbon Calculator - the tool calculates the carbon emissions (CO2) and green infos of any website"
    )

    parser.add_argument(
        "-db",
        "--greenweb",
        type=str,
        help="(Mandatory) - The path of the Green Web Foundation DB (SQL3Lite DB file)",
        required=True,
    )
    parser.add_argument(
        "-lh",
        "--lighthouse",
        type=str,
        help="(Optional) - The path of the Lighthouse tool",
        required=False,
    )
    parser.add_argument("website", type=str, help="The URL to analyze")

    args = parser.parse_args()

    try:
        lighthouse = (
            LighthouseService()
            if not args.lighthouse
            else LighthouseService(args.lighthouse)
        )
        greenweb = GreenWebService(args.greenweb)
        website = args.website

        carbon = CarbonCalculator(lighthouse=lighthouse, greenweb=greenweb)
        carbon.footprint(website)

        print(carbon.to_json())
    except Exception as e:
        print(e)



import json
from .services import LighthouseService, GreenWebService
from .exceptions import CarbonCalculatorException
from datetime import date, datetime, timezone


class PageStatistics(object):
    """Statistics component"""

    KWH_PER_GB = 1.805
    FIRST_TIME_VIEWING_PERCENTAGE = 0.75
    RETURNING_VISITOR_PERCENTAGE = 0.25
    PERCENTAGE_OF_DATA_LOADED_ON_SUBSEQUENT_LOAD = 0.02
    CARBON_PER_KWH_GRID = 475
    CARBON_PER_KWH_RENEWABLE = 33.4
    PERCENTAGE_OF_ENERGY_IN_DATACENTER = 0.1008
    PERCENTAGE_OF_ENERGY_IN_TRANSMISSION_AND_END_USER = 0.8992
    CO2_GRAMS_TO_LITRES = 0.5562

    def __init__(self, url: str, resources: dict, green: bool) -> None:
        self._adjusted_bytes = 0
        self._energy = 0
        self._co2 = 0
        self._co2_grid_grams = 0
        self._co2_grid_litres = 0
        self._co2_renewable_grams = 0
        self._co2_renewable_litres = 0
        self._green = green
        self._resources = resources
        self._transfered_bytes = 0
        self._resources_bytes = 0
        self._url = url
        self._litres = 0
        self._created_at = None

        self._calculate_stats()

    def _calculate_stats(self) -> None:
        """
        It build the statistics (co2, energy, litres)

        Parameters
        ----------
        bytes: int
            The transfered bytes of the website

        """
        self._transfered_bytes = self._resources["transfer_size_bytes"]["total"]
        self._adjusted_bytes = self._adjust_data_transfer(self._transfered_bytes)

        self._resources["transfer_size_bytes"]["total_weighted"] = self._adjusted_bytes
        self._resources_bytes = self._resources["resources_size_bytes"]["total"]

        self._energy = self._energy_consumption(self._adjusted_bytes)

        self._co2_grid_grams = self._get_co2_grid(self._energy)
        self._co2_grid_litres = self._co2_to_litres(self._co2_grid_grams)

        self._co2_renewable_grams = self._get_co2_renewable(self._energy)
        self._co2_renewable_litres = self._co2_to_litres(self._co2_renewable_grams)

        self._co2 = self._co2_renewable_grams if self._green else self._co2_grid_grams
        self._litres = (
            self._co2_renewable_litres if self._green else self._co2_grid_litres
        )

        self._created_at = datetime.now(timezone.utc).astimezone().isoformat()

    def _adjust_data_transfer(self, val: int) -> int:
        return int(
            (val * self.FIRST_TIME_VIEWING_PERCENTAGE)
            + (
                self.PERCENTAGE_OF_DATA_LOADED_ON_SUBSEQUENT_LOAD
                * val
                * self.RETURNING_VISITOR_PERCENTAGE
            )
        )

    def _energy_consumption(self, bytes: int) -> float:
        return bytes * (self.KWH_PER_GB / 1073741824)

    def _get_co2_grid(self, energy: float) -> float:
        return energy * self.CARBON_PER_KWH_GRID

    def _get_co2_renewable(self, energy: float) -> float:
        return (
            (energy * self.PERCENTAGE_OF_ENERGY_IN_DATACENTER)
            * self.CARBON_PER_KWH_RENEWABLE
        ) + (
            (energy * self.PERCENTAGE_OF_ENERGY_IN_TRANSMISSION_AND_END_USER)
            * self.CARBON_PER_KWH_GRID
        )

    def _co2_to_litres(self, co2: float) -> float:
        return co2 * self.CO2_GRAMS_TO_LITRES

    @property
    def created_at(self) -> int:
        return self._created_at

    @property
    def url(self) -> str:
        return self._url

    @property
    def hosting_green(self) -> bool:
        return self._green

    @property
    def co2_grams(self) -> float:
        return self._co2

    @property
    def energy_kWh(self) -> float:
        return self._energy

    @property
    def water_litres(self) -> float:
        return self._litres

    @property
    def resources(self) -> int:
        return self._resources

    @property
    def transfer_size_bytes(self) -> int:
        return self._transfered_bytes

    @property
    def resources_size_bytes(self) -> int:
        return self._resources_bytes


class StatisticsBuilder(object):
    def __init__(
        self, greenweb: GreenWebService, lighthouse: LighthouseService
    ) -> None:
        self._greenweb = greenweb
        self._lighthouse = lighthouse

    def build(self, url: str) -> PageStatistics:
        statistics = None
        try:
            green = self._greenweb.check(url)
            self._lighthouse.analyze(url)
            resources = self._lighthouse.resources
            statistics = PageStatistics(url, resources, green)

            return statistics

        except Exception as e:
            raise CarbonCalculatorException(e)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
import os
import re
import sys

PY3 = sys.version_info[0] == 3
HERE = os.path.dirname(os.path.abspath(__file__))


def get_version():
    filename = os.path.join(HERE, 'carbon', '__init__.py')
    with open(filename) as f:
        contents = f.read()
    pattern = r"^__version__ = \"(.*?)\"$"
    return re.search(pattern, contents, re.MULTILINE).group(1)
    


with open('README.md') as readme_file:
    readme = readme_file.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()


test_requirements = [ ]

setup(
    author="Giovanni Pirrotta",
    author_email='giovanni.pirrotta@gmail.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="Carbon Calculator",
    install_requires=required,
    license="MIT license",
    long_description=readme,
    long_description_content_type='text/markdown',
    py_modules = ["carboncli"],
    platforms='any',
    include_package_data=True, 
    keywords='green,environment,carbon,co2,pollution',
    name='carbon-calculator',
    packages=find_packages(include=['carbon', 'carbon.*']),
    tests_require=test_requirements,
    url='https://github.com/gpirrotta/carbon-calculator',
    version=get_version(),
    zip_safe=False,
    entry_points = {
        'console_scripts': ['carbon-cli=carboncli:main'],
    }
)



import subprocess
import json
import sqlite3
from os.path import isfile, getsize
from .exceptions import CarbonCalculatorException


class LighthouseService(object):
    """Weigh Calculator component

    It collects metrics on websites throgh the external **lighthouse**
    opensource tool

    https://github.com/GoogleChrome/lighthouse
    """

    def __init__(self, lighthouse_path: str = "") -> None:
        self._resources = {}
        self._transfered_bytes = 0
        self._resources_bytes = 0
        self._lighthouse_path = (
            lighthouse_path if lighthouse_path != "" else "lighthouse"
        )
        self._result = {}

    def analyze(self, url) -> None:
        """Collect resources data and calculates the total of transfered bytes

        Parameters
        ----------
        url : str
            The Website to analyze
        """
        try:

            cmd = f"{self._lighthouse_path} --quiet --no-update-notifier --no-enable-error-reporting --output=json --chrome-flags='--headless' {url} --only-audits='network-requests'"

            process = subprocess.Popen(
                cmd,
                shell=True,
                universal_newlines=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            (output, error) = process.communicate()
            if not error:
                self._build_metrics(output)
            else:
                raise CarbonCalculatorException(
                    "Error in Lighthouse tool - the tool must be installed and present in the PATH or the absolute URL must be passed as argument"
                )

        except CarbonCalculatorException as e:
            raise Exception(e)

        finally:
            process.stdout.close()
            process.stderr.close()
            process.terminate()
            process.kill()

    def _build_metrics(self, output):
        mime_types = [
            "html",
            "css",
            "javascript",
            "image",
            "font",
            "audio",
            "video",
            "other",
        ]
        output = json.loads(output)
        items = output["audits"]["network-requests"]["details"]["items"]
        metrics = {}
        metrics["transfer_size_bytes"] = {}
        metrics["transfer_size_bytes"]["total"] = 0
        metrics["transfer_size_bytes"]["total_weighted"] = 0

        metrics["resources_size_bytes"] = {}
        metrics["resources_size_bytes"]["total"] = 0

        for mime in mime_types:
            metrics["transfer_size_bytes"][f"{mime}"] = 0
            metrics["resources_size_bytes"][f"{mime}"] = 0

        for metadata in items:
            found_mime_transfer = False
            if metadata["transferSize"] > 0:
                metrics["transfer_size_bytes"]["total"] += metadata["transferSize"]
                for mime in mime_types:
                    if mime in metadata["mimeType"]:
                        metrics["transfer_size_bytes"][f"{mime}"] += metadata[
                            "transferSize"
                        ]
                        found_mime_transfer = True
                        break
                if not found_mime_transfer:
                    metrics["transfer_size_bytes"]["other"] += metadata["transferSize"]

            found_mime_resource = False
            if metadata["resourceSize"] > 0:
                metrics["resources_size_bytes"]["total"] += metadata["resourceSize"]
                for mime in mime_types:
                    if mime in metadata["mimeType"]:
                        metrics["resources_size_bytes"][f"{mime}"] += metadata[
                            "resourceSize"
                        ]
                        found_mime_resource = True
                        break
                if not found_mime_resource:
                    metrics["resources_size_bytes"]["other"] += metadata["resourceSize"]

        self._resources = metrics

    @property
    def transfered_bytes(self) -> int:
        """The total of bytes transfered"""
        return self._transfered_bytes

    @property
    def resources_bytes(self) -> int:
        return self._resources_bytes

    @property
    def resources(self) -> dict:
        """The collection of the metrics"""
        return self._resources


class GreenWebService(object):
    """The GreenWebService component checks if site is present in the Green Web Dataset

    The dataset can be download here:
        https://www.thegreenwebfoundation.org/green-web-datasets/
    """

    def __init__(self, db_file: str) -> None:
        try:
            if not self._is_valid_sqlite3_db(db_file):
                raise CarbonCalculatorException("The sqlite db is not valid or missing")

            _conn = sqlite3.connect(f"file:{db_file}?mode=ro", uri=True)
            self._cur = _conn.cursor()

        except CarbonCalculatorException as e:
            raise Exception(e)

        except Exception as e:
            raise Exception("Problems in GreenWebService")

    def check(self, site) -> bool:
        """Chek if site is present in the green dataset

        Parameters
        ----------
        site : str
            URL of the website to check


        Returns
        -------
        bool:
            True if website is present in the dataset, False otherwise

        """

        site_cleaned = site.replace("http://", "").replace("https://", "")
        sql = "SELECT EXISTS(SELECT 1 FROM greendomain WHERE url LIKE ?)"

        try:
            self._cur.execute(sql, (f"%{site_cleaned}%",))
            result = self._cur.fetchone()[0]
        except Exception:
            raise Exception
        finally:
            self._cur.close()

        return True if result == 1 else False

    def _is_valid_sqlite3_db(self, filename):
        """
        Check if a file is a SQLite3 database.
        http://stackoverflow.com/questions/12932607/how-to-check-with-python-and-sqlite3-if-one-sqlite-database-file-exists
        """

        if not isfile(filename):
            return False
        if getsize(filename) < 100:  # SQLite database file header is 100 bytes
            return False

        with open(filename, "rb") as fd:
            header = fd.read(100)

        isFileSQLite = False

        if header[0:16] == b"SQLite format 3\000":
            isFileSQLite = True

        return isFileSQLite
